import logging
import os
from typing import Any, Callable

logger = logging.getLogger(__name__)


def get_client_url() -> str:
    """Returns the default URL the client will use if unspecified.

    Can be overridden by the environment variable MODELON_IMPACT_CLIENT_URL.

    """
    url = os.environ.get("MODELON_IMPACT_CLIENT_URL")
    if url is None:
        url = "https://impact.modelon.cloud/"
        logger.warning("No URL for client was specified, will use: {}".format(url))
    return url


def get_client_interactive() -> bool:
    """Returns the default for if client will run interactive or not if unspecified.

    Can be overridden by the environment variable MODELON_IMPACT_CLIENT_INTERACTIVE.

    """
    interactive_env = os.environ.get("MODELON_IMPACT_CLIENT_INTERACTIVE", "false")
    return interactive_env.lower() in ("1", "true")


def get_client_experimental() -> bool:
    """Returns the default for if experimental client methods should be enabled or not.

    Can be overridden by the environment variable IMPACT_PYTHON_CLIENT_EXPERIMENTAL.

    """
    experimental_env = os.environ.get("IMPACT_PYTHON_CLIENT_EXPERIMENTAL", "false")
    return experimental_env.lower() in ("1", "true")


def get_client_experiments_v3_experimental() -> bool:
    """If the client should use the new experimental V3 schema for experiments.

    Can be overridden by the environment variable IMPACT_PYTHON_CLIENT_V3_EXPERIMENTS.

    """
    experimental_env = os.environ.get("IMPACT_PYTHON_CLIENT_V3_EXPERIMENTS", "false")
    return experimental_env.lower() in ("1", "true")


class Experimental:
    def __init__(self, fn: Callable):
        self.fn = fn

    def __set_name__(self, owner: str, name: str) -> None:
        if get_client_experimental():
            setattr(owner, name, self.fn)
        else:
            delattr(owner, name)

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        self.fn(*args, **kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional

from modelon.impact.client.configuration import get_client_experiments_v3_experimental
from modelon.impact.client.options import BaseExecutionOptions

if TYPE_CHECKING:
    from modelon.impact.client.entities.case import Case


def get_options(default_options: Any, options: Optional[Any]) -> Dict[str, Any]:
    return (
        dict(default_options())
        if options is None
        else dict(options)
        if isinstance(options, BaseExecutionOptions)
        else options
    )


def case_to_identifier_dict(case: Case) -> Dict[str, Any]:
    return {
        "caseId": case.id,
        "experimentId": case.experiment_id,
    }


def custom_function_parameters_to_dict(parameter_values: dict[str, Any]) -> Any:
    if get_client_experiments_v3_experimental():
        return [
            {"name": name, "value": value} for name, value in parameter_values.items()
        ]
    else:
        return parameter_values



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any, Dict

logger = logging.getLogger(__name__)


class BaseExperimentDefinition(ABC):
    """Base class for an Experiment definition class."""

    @abstractmethod
    def validate(self) -> None:
        """Validates the modifiers appended to the experiment definition."""

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """Returns the experiment definition as a dictionary."""



from __future__ import annotations

from typing import Any


class Error(Exception):
    pass


class UnsupportedSemanticVersionError(Error):
    pass


class NoAssignedLicenseError(Error):
    pass


class OperationTimeOutError(Error):
    pass


class OperationFailureError(Error):
    @classmethod
    def for_operation(cls, operation_name: str) -> OperationFailureError:
        return cls(
            f"{operation_name} was cancelled before completion! "
            f"Log file generated for cancelled {operation_name} is empty!"
        )


class OperationNotCompleteError(Error):
    @classmethod
    def for_operation(
        cls, operation_name: str, status: Any
    ) -> OperationNotCompleteError:
        return cls(
            f"{operation_name} is still in progress! Status : {status}."
            f" Please call the wait() method on the {operation_name} operation"
            " to wait until completion!"
        )


class OperationCompleteError(Error):
    pass


class IllegalProjectImport(Error):
    pass


class IllegalWorkspaceImport(Error):
    pass


class IllegalWorkspaceExport(Error):
    pass


class IllegalWorkspaceConversion(Error):
    pass


class IllegalContentImport(Error):
    pass


class IllegalFMUImport(Error):
    pass


class ExternalResultUploadError(Error):
    pass


class NoSuchCustomArtifactError(Error):
    pass


class NoAssociatedPublishedWorkspaceError(Error):
    pass


class RemotePublishedWorkspaceLinkError(Error):
    pass



"""HTTP client class."""
from typing import Any, Dict, Optional, Union

from modelon.impact.client.jupyterhub.sal import JupyterContext
from modelon.impact.client.sal.context import Context
from modelon.impact.client.sal.request import (
    RequestCSV,
    RequestJSON,
    RequestMatStream,
    RequestOctetStream,
    RequestText,
    RequestXML,
    RequestZip,
)
from modelon.impact.client.sal.response import (
    CSVResponse,
    JSONResponse,
    MatStreamResponse,
    OctetStreamResponse,
)


class HTTPClient:
    def __init__(self, context: Optional[Union[Context, JupyterContext]] = None):
        self._context = context if context else Context()

    def get_json(self, url: str, headers: Optional[Dict[str, Any]] = None) -> Any:
        return self.get_json_response(url, headers=headers).data

    def get_json_response(
        self, url: str, headers: Optional[Dict[str, Any]] = None
    ) -> JSONResponse:
        request = RequestJSON(self._context, "GET", url, headers=headers)
        return request.execute()

    def get_text(self, url: str) -> str:
        request = RequestText(self._context, "GET", url)
        return request.execute().data

    def get_csv(
        self, url: str, headers: Optional[Dict[str, Any]] = None
    ) -> CSVResponse:
        request = RequestCSV(self._context, "GET", url, headers=headers)
        return request.execute()

    def get_xml(self, url: str, headers: Optional[Dict[str, Any]] = None) -> str:
        request = RequestXML(self._context, "GET", url, headers=headers)
        return request.execute().data

    def get_mat(
        self, url: str, headers: Optional[Dict[str, Any]] = None
    ) -> MatStreamResponse:
        request = RequestMatStream(self._context, "GET", url, headers=headers)
        return request.execute()

    def get_octet_response(self, url: str) -> OctetStreamResponse:
        request = RequestOctetStream(self._context, "GET", url)
        return request.execute()

    def get_zip(self, url: str) -> bytes:
        request = RequestZip(self._context, "GET", url)
        return request.execute().data

    def post_json(
        self,
        url: str,
        body: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ) -> Any:
        request = RequestJSON(self._context, "POST", url, body, files, headers=headers)
        return request.execute().data

    def post_json_no_response_body(
        self, url: str, body: Optional[Dict[str, Any]] = None
    ) -> None:
        RequestJSON(self._context, "POST", url, body).execute()

    def patch_json_no_response_body(
        self, url: str, body: Optional[Dict[str, Any]] = None
    ) -> None:
        RequestJSON(self._context, "PATCH", url, body).execute()

    def delete_json(self, url: str, body: Optional[Dict[str, Any]] = None) -> None:
        RequestJSON(self._context, "DELETE", url, body).execute()

    def put_json_no_response_body(
        self, url: str, body: Optional[Dict[str, Any]] = None
    ) -> None:
        RequestJSON(self._context, "PUT", url, body).execute()

    def put_json(self, url: str, body: Optional[Dict[str, Any]] = None) -> Any:
        request = RequestJSON(self._context, "PUT", url, body)
        return request.execute().data



class Error(Exception):
    pass


class JupyterHubAuthrizationError(Error):
    pass


class NotAJupyterHubUrl(Error):
    pass


class NoJupyterHubServerRunningError(Error):
    pass


class UnknownJupyterHubError(Error):
    pass


class NoJupyterHubTokenError(Error):
    pass



"""This module provides an entry-point to the client APIs."""
from __future__ import annotations

import enum
import logging
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Union

from semantic_version import SimpleSpec, Version  # type: ignore

import modelon.impact.client.jupyterhub
import modelon.impact.client.sal.exceptions
import modelon.impact.client.sal.service
from modelon.impact.client import exceptions
from modelon.impact.client.configuration import get_client_interactive, get_client_url
from modelon.impact.client.credential_manager import CredentialManager
from modelon.impact.client.entities.experiment import Experiment
from modelon.impact.client.entities.model_executable import ModelExecutable
from modelon.impact.client.entities.project import (
    Project,
    ProjectType,
    StorageLocation,
    VcsUri,
)
from modelon.impact.client.entities.workspace import Workspace, WorkspaceDefinition
from modelon.impact.client.operations.experiment import ExperimentOperation
from modelon.impact.client.operations.model_executable import ModelExecutableOperation
from modelon.impact.client.operations.project_import import ProjectImportOperation
from modelon.impact.client.operations.workspace.conversion import (
    WorkspaceConversionOperation,
)
from modelon.impact.client.operations.workspace.imports import WorkspaceImportOperation
from modelon.impact.client.published_workspace_client import PublishedWorkspacesClient
from modelon.impact.client.sal.context import Context
from modelon.impact.client.sal.uri import URI

logger = logging.getLogger(__name__)


@enum.unique
class ExecutionKind(enum.Enum):
    COMPILATION = "COMPILATION"
    EXPERIMENT = "EXPERIMENT"


@dataclass
class Execution:
    kind: ExecutionKind
    workspace_id: str
    id: str

    @classmethod
    def from_dict(cls, execution: Dict[str, Any]) -> Execution:
        kind = ExecutionKind(execution["kind"])
        id = (
            execution["fmu"]["id"]
            if kind == ExecutionKind.COMPILATION
            else execution["experiment"]["id"]
        )
        return cls(
            kind=kind,
            workspace_id=execution["workspace"]["id"],
            id=id,
        )


@dataclass
class Selection:
    entry_id: str
    project: Project

    def to_dict(self) -> Dict[str, Any]:
        return {"id": self.entry_id, "project": {"id": self.project.id}}


@dataclass
class ProjectMatching:
    entry_id: str
    vcs_uri: str
    projects: List[Project]

    def get_selection(self, index: int) -> Selection:
        """Returns the selection given the index of the selected project.

        Args:
            index: The index of the selected project

        Returns:
            A Selection class object.

        Example::

            matchings = client.get_project_matchings(imported_workspace_definition)
            selection = matchings.entries[0].get_selection(index=0)

            client.import_workspace_from_shared_definition(shared_definition).wait()

        """
        return Selection(entry_id=self.entry_id, project=self.projects[index])

    def make_selection_interactive(self) -> Selection:
        """Returns an interactively chosen Selection object.

        Returns:
            A Selection class object.

        Example::

            matchings = client.get_project_matchings(imported_workspace_definition)
            selection = matchings.entries[0].make_selection_interactive()

        """
        if len(self.projects) == 1:
            print(f"Only one project matches the URI {self.vcs_uri}!")
            selection = self.get_selection(index=0)
            print(
                f"Resolving conflict automatically using project with ID: "
                f"{selection.project.id} for repository with URI: {self.vcs_uri}."
            )
            return selection

        print(
            f"\nThe URI {self.vcs_uri} in the workspace definition matched with"
            " multiple projects during import:"
        )
        for i, project in enumerate(self.projects):
            choice = f"[{i}]:"
            vcs_uri = f"{' ' * len(choice)} Project URI: {str(project.vcs_uri)}"
            print("_" * 60)
            print(f"{choice} Project name: {project.definition.name}")
            print(vcs_uri)

        while True:
            try:
                chosen_index = int(input("\nPlease enter one of the above choices:-"))
                selection = self.get_selection(index=chosen_index)
                break
            except (KeyError, ValueError, IndexError):
                allowed_choices = map(str, range(len(self.projects)))
                print(
                    f"Invalid choice. Please select one of {','.join(allowed_choices)}!"
                )

        print(
            f"Resolving conflict using project with ID: {selection.project.id}"
            f" for repository with URI: {self.vcs_uri}."
        )
        return selection


@dataclass
class ProjectMatchings:
    entries: List[ProjectMatching]

    def make_selections_interactive(self) -> List[Selection]:
        """Returns a list of Selection class objects based on interactive input from
        users. As import will fail if there are multiple possible matchings of local
        projects for a project, this method is used to resolve to an unequivocal
        'selection'. The list of preferred Selection objects could be specified during
        workspace import in the
        :obj:`~modelon.impact.client.Client.import_workspace_from_shared_definition`
        method.

        Returns:
            A a list of Selection class objects.

        Example::

            # Interactive workflow
            matchings = client.get_project_matchings(shared_definition)
            selections = matchings.make_selections_interactive()
            imported_workspace = client.import_workspace_from_shared_definition(
                shared_definition, selections
            ).wait()

        """
        return [e.make_selection_interactive() for e in self.entries]


class Client:
    """This Class contains methods to authenticate logins, create new workspaces and
    upload or fetch existing workspaces.

    Args:
        url:
            The URL for Modelon Impact client host. Defaults to the value specified
            by environment variable 'MODELON_IMPACT_CLIENT_URL' if set else uses the URL
            'https://impact.modelon.cloud/'.
        interactive:
            If True the client will prompt for an API key if no other login information
            can be found. An API key entered for this prompt will be saved to disk
            and re-used next time the Client is instantiated. If False no prompt will
            be given if no other login information can be found and login will be done
            as if no API key was given (anonymous login).

            For scripts and notebooks that are running interactively by a user in
            a shell it is recommended to use interactive=True. For scripts or
            applications that are automated or for other reasons won't have a user
            ready to enter an API key it is recommended to use interactive=False.

            Default is False. It is possible to change the default value through
            the environment variable 'MODELON_IMPACT_CLIENT_INTERACTIVE'.
        credential_manager:
            Help class for managing credentials for the Impact server. Default is None
            and then the default credential manager is used.
        context:
            Request contexts to pass data alongside a HTTP request. Default is None and
            then the default context is used.

    Example::

        from modelon.impact.client import Client

        client = Client(url=impact_url)
        client = Client(url=impact_url, interactive=True)

    """

    _SUPPORTED_VERSION_RANGE = ">=4.0.0,<5.0.0"

    def __init__(
        self,
        url: Optional[str] = None,
        interactive: Optional[bool] = None,
        credential_manager: Optional[CredentialManager] = None,
        context: Optional[Context] = None,
        jupyterhub_credential_manager: Optional[CredentialManager] = None,
    ):
        if url is None:
            url = get_client_url()

        if interactive is None:
            interactive = get_client_interactive()

        self._uri = URI(url)
        self._sal = modelon.impact.client.sal.service.Service(self._uri, context)

        if self._sal.is_jupyterhub_url():
            logger.info(
                "API response indicates that the URL '%s' hosts a JupyterHub.",
                str(self._uri),
            )
            self._uri, jupyter_context = modelon.impact.client.jupyterhub.authorize(
                self._uri,
                interactive,
                context,
                jupyterhub_credential_manager,
            )
            self._sal = modelon.impact.client.sal.service.Service(
                self._uri, jupyter_context
            )

        self._validate_compatible_api_version()

        if credential_manager is None:
            help_hint = f"can be generated at {self._uri / 'admin/keys'}"
            help_text = f"Enter Modelon Impact API key ({help_hint}):"
            credential_manager = CredentialManager(interactive_help_text=help_text)
        self._credential_manager = credential_manager

        try:
            api_key = self._authenticate_against_api(interactive)
        except modelon.impact.client.sal.exceptions.HTTPError:
            if interactive:
                logger.warning(
                    "The provided Modelon Impact API key is not valid, "
                    "please enter a new key"
                )
                api_key = self._credential_manager.get_key_from_prompt()
                api_key = self._authenticate_against_api(interactive, api_key=api_key)
            else:
                raise

        self._sal.add_login_retry_with(api_key)

        resp = self._sal.users.get_me()
        if "license" not in resp["data"]:
            raise exceptions.NoAssignedLicenseError(
                "No assigned license during login. Either out of floating Deployment "
                "Add-on licenses or your assigned user license could not be validated"
            )

    def _validate_compatible_api_version(self) -> None:
        try:
            version = self._sal.api_get_metadata()["version"]
        except modelon.impact.client.sal.exceptions.CommunicationError as exce:
            raise modelon.impact.client.sal.exceptions.NoResponseFetchVersionError(
                f"No response from url {self._uri}, "
                "please verify that the URL is correct"
            ) from exce

        if Version(version) not in SimpleSpec(self._SUPPORTED_VERSION_RANGE):
            raise exceptions.UnsupportedSemanticVersionError(
                f"Version '{version}' of the HTTP REST API is not supported, "
                f"must be in the range '{self._SUPPORTED_VERSION_RANGE}'! "
                "Updgrade or downgrade this package to a version "
                f"that supports version '{version}' of the HTTP REST API."
            )

    def _authenticate_against_api(
        self, interactive: bool, api_key: Optional[str] = None
    ) -> Optional[str]:
        if not api_key:
            api_key = self._credential_manager.get_key(interactive=interactive)

        if not api_key:
            logger.warning(
                "No Modelon Impact API key could be found, "
                "will log in as anonymous user. Permissions may be limited"
            )

        self._sal.api_login(api_key=api_key)
        if api_key and interactive:
            # Save the api_key for next time if
            # running interactively and login was successfully
            self._credential_manager.write_key_to_file(api_key)

        return api_key

    def get_workspace(self, workspace_id: str) -> Workspace:
        """Returns a Workspace class object.

        Args:
            workspace_id: The ID of the workspace.

        Returns:
            Workspace class object.

        Example::

            client.get_workspace('my_workspace')

        """
        resp = self._sal.workspace.workspace_get(workspace_id, size_info=False)
        return Workspace(resp["id"], self._sal)

    def get_workspace_by_name(self, workspace_name: str) -> List[Workspace]:
        """Returns a list of Workspace class objects with the given name.

        Args:
            workspace_name: The name of the workspace.

        Returns:
            A list of workspace class objects with the given name.

        Example::

            workspaces = client.get_workspace_by_name('TestWorkspace')

        """
        resp = self._sal.workspace.workspaces_get()
        return [
            Workspace(item["id"], self._sal)
            for item in resp["data"]["items"]
            if item["definition"]["name"] == workspace_name
        ]

    def get_workspaces(
        self,
        only_app_mode: bool = False,
        name: Optional[str] = None,
        sharing_id: Optional[str] = None,
    ) -> List[Workspace]:
        """Returns a list of Workspace class object.

        Returns:
            A list of Workspace class objects.

        Example::

            workspaces = client.get_workspaces()

        """
        resp = self._sal.workspace.workspaces_get(only_app_mode, name, sharing_id)
        return [Workspace(item["id"], self._sal) for item in resp["data"]["items"]]

    def convert_workspace(
        self, workspace_id: str, backup_name: Optional[str] = None
    ) -> WorkspaceConversionOperation:
        """Converts a workspace of an old version up to the new version the server is
        using.

        Args:
            workspace_id: The ID of the workspace to convert to the latest
            backup_name: If given then a backup will be created with this
            name. Defaults to None.

        Returns:
            The workspace conversion operation

        Example::

            workspace = client.convert_workspace(workspace_id, backup_name='old save')
            .wait()

        """
        resp = self._sal.workspace.workspace_conversion_setup(workspace_id, backup_name)
        return WorkspaceConversionOperation[Workspace](
            resp["data"]["location"], self._sal, Workspace.from_conversion_operation
        )

    def get_project(self, project_id: str, vcs_info: bool = True) -> Project:
        """Returns a project class object.

        Args:
            project_id: The id of the project.
            vcs_info: If True, the versioning details are returned for the
                project(if under version control).

        Returns:
            Project class objects.

        Example::

            client.get_project('hcbhsb11313321')

        """
        resp = self._sal.project.project_get(
            project_id, vcs_info=vcs_info, size_info=False
        )
        return Project(
            resp["id"],
            resp["projectType"],
            resp["storageLocation"],
            VcsUri.from_dict(resp["vcsUri"]) if resp.get("vcsUri") else None,
            self._sal,
        )

    def get_projects(
        self,
        vcs_info: bool = True,
        project_type: Optional[ProjectType] = None,
        storage_location: Optional[StorageLocation] = None,
    ) -> List[Project]:
        """Returns a list of project class object.

        Args:
            vcs_info: If True, the versioning details are returned for the
                projects under version control.
            project_type: Used to filter so only projects of a specified ProjectType
                are returned. If not given all project types are returned.
            storage_location: Used to filter so only projects of a specified
                StorageLocation are returned. If not given all project in all
                storage locations are returned.

        Returns:
            A list of Project class objects.

        Example::

            from modelon.impact.client import ProjectType, StorageLocation

            client.get_projects()
            client.get_projects(
                project_type=ProjectType.LOCAL,
                storage_location=StorageLocation.USERSPACE
            )

        """
        resp = self._sal.project.projects_get(
            vcs_info=vcs_info,
            project_type=project_type,
            storage_location=storage_location,
        )
        return [
            Project(
                item["id"],
                item["projectType"],
                item["storageLocation"],
                VcsUri.from_dict(item["vcsUri"]) if item.get("vcsUri") else None,
                self._sal,
            )
            for item in resp["data"]["items"]
        ]

    def create_workspace(self, workspace_name: str) -> Workspace:
        """Creates and returns a Workspace. Returns a workspace class object.

        Args:
            workspace_name: The name of the workspace to create.

        Returns:
            The created workspace class object.

        Example::

            workspace = client.create_workspace('my_workspace')

        """
        resp = self._sal.workspace.workspace_create(workspace_name)
        return Workspace(resp["id"], self._sal)

    def upload_workspace(self, path_to_workspace: str) -> Workspace:
        """Imports a Workspace from a compressed(.zip) workspace file. Returns the
        workspace class object of the imported workspace. Similar to
        :obj:`~modelon.impact.client.Client.import_workspace_from_zip`, but does the
        import in one go.

        Args:
            path_to_workspace: The path for the compressed workspace(.zip)
            to be uploaded.

        Returns:
            Workspace class object.

        Example::

            workspace = client.upload_workspace(path_to_workspace)

        """
        return self.import_workspace_from_zip(path_to_workspace).wait()

    def import_workspace_from_zip(
        self, path_to_workspace: str
    ) -> WorkspaceImportOperation:
        """Imports a Workspace from a compressed(.zip) workspace file. Similar to
        :obj:`~modelon.impact.client.Client.upload_workspace`, but gives more control
        for getting the workspace async. Returns an WorkspaceImportOperation class
        object.

        Args:
            path_to_workspace: The path for the compressed
            workspace(.zip) to be uploaded.

        Returns:
            A WorkspaceImportOperation class object.

        Example::

            client.import_workspace_from_zip(path_to_workspace).wait()

        """
        resp = self._sal.workspace.import_from_zip(path_to_workspace)
        return WorkspaceImportOperation[Workspace](
            resp["data"]["location"], self._sal, Workspace.from_import_operation
        )

    def import_workspace_from_shared_definition(
        self,
        shared_definition: WorkspaceDefinition,
        selections: Optional[List[Selection]] = None,
    ) -> WorkspaceImportOperation:
        """Imports a Workspace from a shared workspace definition.

        Args:
            shared_definition: The workspace definition for the shared workspace
            selection: Optional list of Selection class objects. This can be specified
            if there are multiple existing projects with the same version control URI.

        Returns:
            A WorkspaceImportOperation class object.

        Example::

            # Import with no conflicts
            client.import_workspace_from_shared_definition(shared_definition).wait()

            # Import with conflicts(Multiple existing projects matches the URI)
            # Programatic workflow to resolve conflicts

            matchings = client.get_project_matchings(shared_definition)
            # Assume the first in list of matchings is good enough:
            selections = [entry.get_selection(index=0) for entry in matchings.entries]
            imported_workspace = client.import_workspace_from_shared_definition(
                shared_definition, selections
            ).wait()

            # Interactive workflow
            matchings = client.get_project_matchings(shared_definition)
            selections = matchings.make_selections_interactive()
            imported_workspace = client.import_workspace_from_shared_definition(
                shared_definition, selections
            ).wait()

        """
        resp = self._sal.workspace.import_from_shared_definition(
            {"definition": shared_definition.to_dict()},
            selected_matchings=[selection.to_dict() for selection in selections]
            if selections
            else None,
        )
        return WorkspaceImportOperation[Workspace](
            resp["data"]["location"], self._sal, Workspace.from_import_operation
        )

    def get_project_matchings(
        self, shared_definition: WorkspaceDefinition
    ) -> ProjectMatchings:
        """Returns all projects matchings that would happen during a workspace import.
        As import will fail if there are multiple possible matchings of local projects
        for a project, this method is used to get these matchings which can be resolved
        to an unequivocal 'selection'. Selections are used as (optional) input to the
        :obj:`~modelon.impact.client.Client.import_workspace_from_shared_definition`
        method.

        Args:
            shared_definition: The workspace definition for the shared workspace

        Returns:
            A ProjectMatchings class object.

        Example::

            # Import with conflicts(Multiple existing projects matches the URI)
            # Programatic workflow to resolve conflicts
            matchings = client.get_project_matchings(shared_definition)
            # Assume the first in list of matchings is good enough:
            selections = [entry.get_selection(index=0) for entry in matchings.entries]
            imported_workspace = client.import_from_shared_definition(
                shared_definition, selections
            ).wait()

            # Interactive workflow
            matchings = client.get_project_matchings(shared_definition)
            selections = matchings.make_selections_interactive()
            imported_workspace = client.import_workspace_from_shared_definition(
                shared_definition, selections
            ).wait()

        """
        project_matchings = []
        matchings = self._sal.workspace.get_project_matchings(
            {"definition": shared_definition.to_dict()}
        )["data"]["vcs"]
        for entry in matchings:
            projects = [
                Project(
                    project["id"],
                    project["projectType"],
                    project["storageLocation"],
                    VcsUri.from_dict(project["vcsUri"]),
                    self._sal,
                )
                for project in entry["projects"]
            ]

            project_matchings.append(
                ProjectMatching(
                    entry["entryId"], str(VcsUri.from_dict(entry["uri"])), projects
                )
            )
        return ProjectMatchings(project_matchings)

    def import_project_from_zip(self, path_to_project: str) -> ProjectImportOperation:
        """Imports a Project from a compressed(.zip) project file. Returns the project
        class object.

        Args:
            path_to_project: The path for the compressed project(.zip)
            to be uploaded.

        Returns:
            A ProjectImportOperation class object.

        Example::

            client.import_project_from_zip(path_to_project).wait()

        """
        resp = self._sal.project.import_from_zip(path_to_project)
        return ProjectImportOperation[Project](
            resp["data"]["location"], self._sal, Project.from_operation
        )

    def _experiment_operation_from_execution(
        self, execution: Execution
    ) -> ExperimentOperation:
        return ExperimentOperation[Experiment](
            execution.workspace_id,
            execution.id,
            self._sal,
            Experiment.from_operation,
        )

    def _model_executable_operation_from_execution(
        self, execution: Execution
    ) -> ModelExecutableOperation:
        return ModelExecutableOperation[ModelExecutable](
            execution.workspace_id,
            execution.id,
            self._sal,
            ModelExecutable.from_operation,
        )

    def _operation_from_execution(
        self, execution: Execution
    ) -> Optional[Union[ExperimentOperation, ModelExecutableOperation]]:
        if execution.kind == ExecutionKind.EXPERIMENT:
            return self._experiment_operation_from_execution(execution)
        return self._model_executable_operation_from_execution(execution)

    def get_executions(
        self, workspace_id: Optional[str] = None
    ) -> Iterable[Optional[Union[ExperimentOperation, ModelExecutableOperation]]]:
        """Yields running/active executions.

        Args:
            workspace_id: The id of the workspace.

        Yields:
            An ExperimentOperation or a ModelExecutableOperation class.

        Example::

            list(client.get_executions())

        """
        executions = self._sal.get_executions()["data"]["items"]
        for execution in executions:
            execution = Execution.from_dict(execution)
            if workspace_id and execution.workspace_id != workspace_id:
                continue
            yield self._operation_from_execution(execution)

    def get_published_workspaces_client(self) -> PublishedWorkspacesClient:
        """Return the PublishedWorkspacesClient class object.

        Example::

            pw_client = client.get_published_workspaces_client()
            pw_client.get_by_id("2h98hciwsniucwincj")

        """
        return PublishedWorkspacesClient(self._sal)

    def get_me(self) -> User:
        """Return the User class object for the logged in user.

        Example::

            user = client.get_me()
            user_tenant_group_name = user.tenant.group_name

        """
        user = self._sal.users.get_me()["data"]
        return User.from_dict(user)


@dataclass
class Tenant:
    id: str
    group_name: str

    @classmethod
    def from_dict(cls, tenant_data: Dict[str, Any]) -> Tenant:
        return Tenant(id=tenant_data["id"], group_name=tenant_data["groupName"])


@dataclass
class User:
    id: str
    username: str
    firstname: str
    last_name: str
    email: str
    license: str
    roles: List[str]
    tenant: Tenant

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> User:
        return User(
            id=data["id"],
            username=data["username"],
            firstname=data["firstName"],
            last_name=data["lastName"],
            email=data["email"],
            license=data["license"],
            roles=data["roles"],
            tenant=Tenant.from_dict(tenant_data=data["tenant"]),
        )



"""Project service module."""
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from modelon.impact.client.sal.http import HTTPClient
from modelon.impact.client.sal.uri import URI

if TYPE_CHECKING:
    from modelon.impact.client.entities.project import ProjectType, StorageLocation


class ProjectService:
    def __init__(self, uri: URI, http_client: HTTPClient):
        self._base_uri = uri
        self._http_client = http_client

    def projects_get(
        self,
        vcs_info: bool,
        project_type: Optional[ProjectType] = None,
        storage_location: Optional[StorageLocation] = None,
    ) -> Dict[str, Any]:
        prj_type = "&type=" + project_type.value if project_type else ""
        stg_loc = (
            "&storageLocation=" + storage_location.value if storage_location else ""
        )
        url = (
            self._base_uri / f"api/projects?vcsInfo={vcs_info}{prj_type}{stg_loc}"
        ).resolve()
        return self._http_client.get_json(url)

    def project_get(
        self, project_id: str, vcs_info: bool, size_info: bool
    ) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/projects/{project_id}?vcsInfo={vcs_info}&sizeInfo={size_info}"
        ).resolve()
        return self._http_client.get_json(url)

    def project_delete(self, project_id: str) -> None:
        url = (self._base_uri / f"api/projects/{project_id}").resolve()
        self._http_client.delete_json(url)

    def project_put(self, project_id: str, project_data: Dict[str, Any]) -> None:
        url = (self._base_uri / f"api/projects/{project_id}").resolve()
        self._http_client.put_json(url, body=project_data)

    def project_options_get(
        self, project_id: str, workspace_id: str, custom_function: str
    ) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/projects/{project_id}/custom-functions/"
            f"{custom_function}/options"
        ).resolve()
        return self._http_client.get_json(url)

    def project_default_options_get(
        self, workspace_id: str, custom_function: str
    ) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/custom-functions/{custom_function}"
            "/default-options"
        ).resolve()
        return self._http_client.get_json(url)

    def project_content_delete(self, project_id: str, content_id: str) -> None:
        url = (
            self._base_uri / f"api/projects/{project_id}/content/{content_id}"
        ).resolve()
        self._http_client.delete_json(url)

    def project_content_upload(
        self, path_to_result: str, project_id: str, content_type: str
    ) -> Dict[str, Any]:
        url = (self._base_uri / f"/api/projects/{project_id}/content-imports").resolve()
        with open(path_to_result, "rb") as f:
            multipart_form_data = {
                "file": f,
                "options": json.dumps({"contentType": content_type}),
            }
            return self._http_client.post_json(url, files=multipart_form_data)

    def project_content_get(self, project_id: str, content_id: str) -> Dict[str, Any]:
        url = (
            self._base_uri / f"/api/projects/{project_id}/content/{content_id}"
        ).resolve()
        return self._http_client.get_json(url)

    def fmu_import(
        self,
        project_id: str,
        content_id: str,
        fmu_path: str,
        class_name: str,
        overwrite: bool = False,
        include_patterns: Optional[Union[str, List[str]]] = None,
        exclude_patterns: Optional[Union[str, List[str]]] = None,
        top_level_inputs: Optional[Union[str, List[str]]] = None,
        step_size: float = 0.0,
    ) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/projects/{project_id}/content/{content_id}/fmu-imports"
        ).resolve()
        options = {
            "className": class_name,
            "overwrite": overwrite,
            "stepSize": step_size,
        }

        if include_patterns:
            options["includePatterns"] = include_patterns
        if exclude_patterns:
            options["excludePatterns"] = exclude_patterns
        if top_level_inputs:
            options["topLevelInputs"] = top_level_inputs

        with open(fmu_path, "rb") as f:
            multipart_form_data = {
                "file": f,
                "options": json.dumps(options),
            }
            return self._http_client.post_json(url, files=multipart_form_data)

    def import_from_zip(self, path_to_project: str) -> Dict[str, Any]:
        url = (self._base_uri / "api/project-imports").resolve()
        with open(path_to_project, "rb") as f:
            return self._http_client.post_json(url, files={"file": f})



import json.decoder
from typing import Optional

import requests

from modelon.impact.client.jupyterhub import exceptions
from modelon.impact.client.sal.context import Context
from modelon.impact.client.sal.uri import URI


class JupyterContext:
    __slots__ = ["_base", "_token"]

    def __init__(self, base: Optional[Context] = None):
        self._base = base if base else Context()
        self._token: Optional[str] = None

    @property
    def session(self) -> requests.Session:
        return self._base.session

    @property
    def token(self) -> Optional[str]:
        return self._token

    @token.setter
    def token(self, token: str) -> None:
        self._token = token
        self.session.headers.update({"Authorization": "token %s" % token})


class JupyterUser:
    def __init__(self, user_id: str, server: Optional[str]):
        self.id = user_id
        self._server = server

    def server_running(self) -> bool:
        return self._server is not None

    def impact_server_uri(self, jupyterhub_uri: URI) -> URI:
        return jupyterhub_uri / f"user/{self.id}/impact"


class JupyterHubService:
    @classmethod
    def get_user_data(
        cls, uri: URI, context: JupyterContext, server: Optional[str] = None
    ) -> JupyterUser:
        auth_token_url = (
            uri / f"hub/api/authorizations/token/{context.token}"
        ).resolve()

        try:
            user_response = context.session.get(auth_token_url)
        except requests.exceptions.RequestException as e:
            raise exceptions.NotAJupyterHubUrl(
                f"Did not get a response from {uri}, is the URL correct?"
            ) from e

        if not user_response.ok:
            if user_response.status_code == 403:
                raise exceptions.JupyterHubAuthrizationError(
                    "Could not authorize against JupyterHub, is the token correct?"
                )
            elif user_response.status_code == 404:
                raise exceptions.NotAJupyterHubUrl(
                    f"Missing resource '{auth_token_url}'. Possible errors are "
                    f"that the URL '{uri}' don't points to a JupyterHub or "
                    "that the user of the API key as been deleted."
                )
            else:
                raise exceptions.UnknownJupyterHubError(
                    "Something went wrong calling JupyterHub. Got the response "
                    f"'{user_response}' with content '{str(user_response.content)}'"
                )

        try:
            user_data = user_response.json()
            return JupyterUser(user_data["name"], user_data.get("server", server))
        except (KeyError, json.decoder.JSONDecodeError) as e:
            raise exceptions.NotAJupyterHubUrl(
                "User response data is not correct, "
                f"is the URL '{uri}' for a JupyterHub?"
            ) from e



"""Request class."""
import logging
from typing import Any, Callable, Dict, Optional

import requests

from modelon.impact.client.sal import exceptions
from modelon.impact.client.sal.response import (
    CSVResponse,
    JSONResponse,
    MatStreamResponse,
    OctetStreamResponse,
    TextResponse,
    XMLResponse,
    ZIPResponse,
)

logger = logging.getLogger(__name__)


class Request:
    def __init__(
        self,
        context: Any,
        method: str,
        url: str,
        request_type: Callable,
        body: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ):
        self.context = context
        self.method = method
        self.url = url
        self.body = body
        self.files = files
        self.request_type = request_type
        self.headers = headers

    def execute(self, check_return: bool = True) -> Any:
        try:
            if self.method == "POST":
                logger.debug("POST with JSON body: {}".format(self.body))
                resp = self.context.session.post(
                    self.url, json=self.body, files=self.files, headers=self.headers
                )
            elif self.method == "GET":
                resp = self.context.session.get(self.url, headers=self.headers)
            elif self.method == "PUT":
                resp = self.context.session.put(
                    self.url, json=self.body, headers=self.headers
                )
            elif self.method == "PATCH":
                resp = self.context.session.patch(
                    self.url,
                    json=self.body,
                    headers=self.headers,
                )
            elif self.method == "DELETE":
                resp = self.context.session.delete(self.url, json=self.body)
            else:
                raise NotImplementedError()
        except requests.exceptions.SSLError as exce:
            raise exceptions.SSLError(
                "SSL error, could not verify connection. Please check that "
                "certificates are setup correctly for the Modelon Impact server"
            ) from exce
        except requests.exceptions.RequestException as exce:
            raise exceptions.CommunicationError(
                "Communication when doing a request failed"
            ) from exce

        resp = self.request_type(resp)
        if check_return and not resp.ok:
            raise exceptions.HTTPError(resp.error.message, resp.status_code)

        return resp


class RequestJSON(Request):
    def __init__(
        self,
        context: Any,
        method: str,
        url: str,
        body: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(context, method, url, JSONResponse, body, files, headers)


class RequestZip(Request):
    def __init__(
        self,
        context: Any,
        method: str,
        url: str,
        body: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(context, method, url, ZIPResponse, body, files, headers)


class RequestText(Request):
    def __init__(
        self,
        context: Any,
        method: str,
        url: str,
        body: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(context, method, url, TextResponse, body, files, headers)


class RequestXML(Request):
    def __init__(
        self,
        context: Any,
        method: str,
        url: str,
        body: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(context, method, url, XMLResponse, body, files, headers)


class RequestCSV(Request):
    def __init__(
        self,
        context: Any,
        method: str,
        url: str,
        body: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(context, method, url, CSVResponse, body, files, headers)


class RequestOctetStream(Request):
    def __init__(
        self,
        context: Any,
        method: str,
        url: str,
        body: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            context, method, url, OctetStreamResponse, body, files, headers
        )


class RequestMatStream(Request):
    def __init__(
        self,
        context: Any,
        method: str,
        url: str,
        body: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(context, method, url, MatStreamResponse, body, files, headers)



from __future__ import annotations

import logging
import os
import tempfile
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from modelon.impact.client import exceptions
from modelon.impact.client.entities.asserts import assert_successful_operation
from modelon.impact.client.entities.interfaces.model_executable import (
    ModelExecutableInterface,
)
from modelon.impact.client.entities.log import Log
from modelon.impact.client.entities.status import ModelExecutableStatus
from modelon.impact.client.experiment_definition.fmu_based import (
    SimpleFMUExperimentDefinition,
)
from modelon.impact.client.operations.model_executable import (
    CachedModelExecutableOperation,
    ModelExecutableOperation,
)

if TYPE_CHECKING:
    from modelon.impact.client.entities.custom_function import CustomFunction
    from modelon.impact.client.operations.base import BaseOperation
    from modelon.impact.client.options import SimulationOptions, SolverOptions
    from modelon.impact.client.sal.service import Service

    SimulationOptionsOrDict = Union[SimulationOptions, Dict[str, Any]]
    SolverOptionsOrDict = Union[SolverOptions, Dict[str, Any]]

logger = logging.getLogger(__name__)


def _assert_compilation_is_complete(
    status: ModelExecutableStatus, operation_name: str = "Operation"
) -> None:
    if status == ModelExecutableStatus.NOTSTARTED:
        raise exceptions.OperationNotCompleteError.for_operation(operation_name, status)
    elif status == ModelExecutableStatus.CANCELLED:
        raise exceptions.OperationFailureError.for_operation(operation_name)


class ModelDescription(str):
    """ModelDescription class inheriting from string object."""

    def show(self) -> None:
        """Prints the formatted xml."""
        print(self)

    def download(self, path: Optional[str] = None) -> str:
        """Downloads the formatted xml.

        Args:
            path: The local path to store the downloaded model description.
                Default: None. If no path is given, model description will
                be downloaded in a temporary directory.

        Returns:
            Local path to the model description file.

        Example::

            model_description = fmu.get_model_description()
            model_description.show()
            model_description.download()

        """
        if path is None:
            path = os.path.join(tempfile.gettempdir(), "impact-downloads")
        os.makedirs(path, exist_ok=True)
        artifact_path = os.path.join(path, "modelDescription.xml")
        with open(artifact_path, mode="w", encoding="utf-8") as f:
            f.write(self)
        return artifact_path


class ModelExecutableRunInfo:
    def __init__(self, status: ModelExecutableStatus, errors: List[str]):
        self._status = status
        self._errors = errors

    @property
    def status(self) -> ModelExecutableStatus:
        """Status info for a Model-Executable.

        .. Note:: For tracing status changes of a running
            compilation, the ModelExecutableOperation status
            property must be used and not ModelExecutable
            entities run info(ModelExecutableRunInfo) status
            property. See example for usage.

        Example::

            # To get the case entity run info status
            status = fmu.run_info.status

            # To get the status updates of a running compilation
            compile_ops = = model.compile()
            status = compile_ops.status

        """
        return self._status

    @property
    def errors(self) -> List[str]:
        """A list of errors.

        Is empty unless 'status' attribute is 'FAILED'

        """
        return self._errors


class ModelExecutableInput:
    """Class containing ModelExecutable input."""

    __slots__ = ["_data"]

    def __init__(self, data: Dict[str, Any]):
        self._data = data

    @property
    def class_name(self) -> str:
        """Returns the model class name."""
        return self._data["class_name"]

    @property
    def compiler_log_level(self) -> str:
        """Returns the compiler log level."""
        return self._data["compiler_log_level"]

    @property
    def compiler_options(self) -> Dict[str, Any]:
        """Returns the compiler options dict."""
        return self._data["compiler_options"]

    @property
    def runtime_options(self) -> Dict[str, Any]:
        """Returns the runtime options dict."""
        return self._data["runtime_options"]

    @property
    def fmi_target(self) -> str:
        """Returns the FMI target."""
        return self._data["fmi_target"]

    @property
    def fmi_version(self) -> str:
        """Returns the FMI version."""
        return self._data["fmi_version"]

    @property
    def platform(self) -> str:
        """Returns the platform the FMU was compiled for."""
        return self._data["platform"]


class ModelExecutable(ModelExecutableInterface):
    """Class containing ModelExecutable functionalities."""

    def __init__(
        self,
        workspace_id: str,
        fmu_id: str,
        service: Service,
        info: Optional[Dict[str, Any]] = None,
        modifiers: Optional[Dict[str, Any]] = None,
    ):
        self._fmu_id = fmu_id
        self._workspace_id = workspace_id
        self._sal = service
        self._info = info
        self._modifiers = modifiers

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, ModelExecutable) and obj._fmu_id == self._fmu_id

    def __repr__(self) -> str:
        return f"FMU with id '{self._fmu_id}'"

    def __hash__(self) -> int:
        return self._fmu_id.__hash__()

    def _variable_modifiers(self) -> Dict[str, Any]:
        return {} if self._modifiers is None else self._modifiers

    def _get_info(self, cached: bool = True) -> Dict[str, Any]:
        if not cached or self._info is None:
            self._info = self._sal.workspace.fmu_get(self._workspace_id, self._fmu_id)

        return self._info

    @property
    def id(self) -> str:
        """FMU id."""
        return self._fmu_id

    @property
    def info(self) -> Dict[str, Any]:
        """Deprecated, use 'run_info' attribute."""
        logger.warning("This attribute is deprectated, use 'run_info' instead")
        return self._get_info(cached=False)

    @property
    def run_info(self) -> ModelExecutableRunInfo:
        """Compilation run information."""
        run_info = self._get_info(cached=False)["run_info"]
        status = ModelExecutableStatus(run_info["status"])
        errors = run_info.get("errors", [])
        return ModelExecutableRunInfo(status, errors)

    @property
    def metadata(self) -> Dict[str, Any]:
        """FMU metadata.

        Returns the 'iteration_variable_count' and 'residual_variable_count' only for
        steady state model compiled as an FMU

        """
        assert_successful_operation(self.is_successful(), "Compilation")
        parameter_state = {"parameterState": self._variable_modifiers()}
        return self._sal.model_executable.ss_fmu_metadata_get(
            self._workspace_id, self._fmu_id, parameter_state
        )

    @property
    def input(self) -> ModelExecutableInput:
        """ModelExecutable input attributes.

        Example::

            compiled_platform = fmu.input.platform

            help(fmu.input.platform) # See help for attribute
            dir(fmu.input) # See nested attributes

        """
        return ModelExecutableInput(self._get_info()["input"])

    def get_class_name(self) -> str:
        """Return the model class name."""
        return self._get_info()["input"]["class_name"]

    def is_successful(self) -> bool:
        """Returns True if the model has compiled successfully. Use the 'run_info'
        attribute to get more info.

        Returns:
            True, if model has compiled successfully. False, if compilation
            process has failed, is running or is cancelled.

        Example::

            fmu.is_successful()

        """
        return self.run_info.status == ModelExecutableStatus.SUCCESSFUL

    def get_log(self) -> Log:
        """Returns the compilation log object.

        Returns:
            The compilation log object.

        Raises:
            OperationNotCompleteError if compilation process is in progress.
            OperationFailureError if compilation process was cancelled.

        Example::

            log = fmu.get_log()
            log.show()

        """
        _assert_compilation_is_complete(self.run_info.status, "Compilation")
        return Log(
            self._sal.model_executable.compile_log(self._workspace_id, self._fmu_id)
        )

    def get_model_description(self) -> ModelDescription:
        """Returns the model description object.

        Returns:
            The ModelDescription class object.

        Example::

            model_description = fmu.get_model_description()

            # Print the formatted xml
            model_description.show()

            # Download the formatted xml
            model_description.download()

            # Parse the xml
            from xml.etree import ElementTree

            tree = ElementTree.fromstring(model_description)
            model_variables =tree.find('ModelVariables')
            variable_names = [child.attrib.get('name') for child in model_variables]

        """
        return ModelDescription(
            self._sal.model_executable.model_description_get(
                self._workspace_id, self._fmu_id
            )
        )

    def delete(self) -> None:
        """Deletes an FMU.

        Example::

            fmu.delete()

        """
        self._sal.model_executable.fmu_delete(self._workspace_id, self._fmu_id)

    def get_settable_parameters(self) -> List[str]:
        """Returns a list of settable parameters for the FMU.

        Returns:
            A list of parameters that can be set on the FMU.

        Raises:
            OperationNotCompleteError if compilation process is in progress.
            OperationFailureError if compilation process has failed or was cancelled.

        Example::

            fmu.get_settable_parameters()

        """
        assert_successful_operation(self.is_successful(), "Compilation")
        return self._sal.model_executable.settable_parameters_get(
            self._workspace_id, self._fmu_id
        )

    def new_experiment_definition(
        self,
        custom_function: CustomFunction,
        solver_options: Optional[SolverOptionsOrDict] = None,
        simulation_options: Optional[SimulationOptionsOrDict] = None,
        simulation_log_level: str = "WARNING",
    ) -> SimpleFMUExperimentDefinition:
        """Returns a new experiment definition using this FMU.

        Args:
            custom_function: The custom function to use for this experiment.
            solver_options: The solver options to use for this experiment. By default
                the options is set to None, which means the default options for the
                custom_function input is used.
            simulation_options: The simulation_options to use for this experiment.
                By default the options is set to None, which means the default options
                for the custom_function input is used.
            simulation_log_level: Simulation log level for this experiment. Default
                is 'WARNING'.

        Example::

            fmu = model.compile().wait()
            dynamic = workspace.get_custom_function('dynamic')
            solver_options = {'atol':1e-8}
            simulation_options = dynamic.get_simulation_options().
            with_values(ncp=500)
            experiment_definition = fmu.new_experiment_definition(
                dynamic, solver_options, simulation_options)
            experiment = workspace.execute(experiment_definition).wait()

        """
        return SimpleFMUExperimentDefinition(
            self,
            custom_function,
            solver_options,
            simulation_options,
            simulation_log_level,
        )

    def download(self, path: Optional[str] = None) -> str:
        """Downloads an FMU binary that is compiled. Returns the local path to the
        downloaded FMU archive.

        Args:
            path: The local path to store the downloaded FMU. Default: None.
                If no path is given, FMU will be downloaded in a temporary directory.

        Returns:
            Local path to the downloaded FMU.

        Raises:
            OperationNotCompleteError if compilation process is in progress.
            OperationFailureError if compilation process has failed or was cancelled.

        Example::

            fmu =  model.compile().wait()
            if fmu.is_successful():
                fmu_path = fmu.download()
                fmu_path = fmu.download('~/Downloads')

        """
        assert_successful_operation(self.is_successful(), "Compilation")
        data = self._sal.workspace.fmu_download(self._workspace_id, self._fmu_id)
        if path is None:
            path = os.path.join(tempfile.gettempdir(), "impact-downloads")
        os.makedirs(path, exist_ok=True)
        fmu_path = os.path.join(path, self._fmu_id + ".fmu")
        with open(fmu_path, "wb") as f:
            f.write(data)
        return fmu_path

    @classmethod
    def from_operation(
        cls, operation: BaseOperation[ModelExecutable], **kwargs: Any
    ) -> ModelExecutable:
        assert isinstance(
            operation, (ModelExecutableOperation, CachedModelExecutableOperation)
        )
        return cls(**kwargs, service=operation._sal)



"""URI class."""
from __future__ import annotations

import sys
import urllib.parse
from typing import Any


class URI:
    def __init__(self, content: str):
        # If running on Windows you can get a lot of overhead using 'localhost'
        if sys.platform.startswith("win32") and content.startswith("http://localhost:"):
            content = content.replace("http://localhost:", "http://127.0.0.1:")

        self.content = content

    def resolve(self, **kwargs: Any) -> str:
        return self.content.format(**kwargs)

    def _with_path(self, path: str) -> URI:
        return URI(urllib.parse.urljoin(self.content + "/", path.lstrip("/")))

    def __floordiv__(self, other: str) -> URI:
        return self._with_path(other)

    def __truediv__(self, other: str) -> URI:
        return self._with_path(other)

    def __repr__(self) -> str:
        return self.content



"""Workspace service module."""
from typing import Any, Dict, List, Optional

from modelon.impact.client.configuration import get_client_experiments_v3_experimental
from modelon.impact.client.sal.http import HTTPClient
from modelon.impact.client.sal.uri import URI


class WorkspaceService:
    def __init__(self, uri: URI, http_client: HTTPClient):
        self._base_uri = uri
        self._http_client = http_client
        if get_client_experiments_v3_experimental():
            self._experiment_schema = "application/vnd.impact.experiment.v3+json"
        else:
            self._experiment_schema = "application/vnd.impact.experiment.v2+json"

    def workspace_create(self, name: str) -> Dict[str, Any]:
        url = (self._base_uri / "api/workspaces").resolve()
        return self._http_client.post_json(url, body={"new": {"name": name}})

    def workspace_delete(self, workspace_id: str) -> None:
        url = (self._base_uri / f"api/workspaces/{workspace_id}").resolve()
        self._http_client.delete_json(url)

    def workspace_get(self, workspace_id: str, size_info: bool) -> Dict[str, Any]:
        url = (
            self._base_uri / f"api/workspaces/{workspace_id}?sizeInfo={size_info}"
        ).resolve()
        return self._http_client.get_json(url)

    def workspaces_get(
        self,
        only_app_mode: bool = False,
        name: Optional[str] = None,
        sharing_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        query = {}
        if name:
            query["name"] = name
        if sharing_id:
            query["sharingId"] = sharing_id
        if only_app_mode:
            query["onlyAppMode"] = str(only_app_mode)
        query_args = "&".join(f"{key}={value}" for key, value in query.items())
        url = (self._base_uri / f"api/workspaces?{query_args}").resolve()
        return self._http_client.get_json(url)

    def projects_get(
        self, workspace_id: str, vcs_info: bool, include_disabled: bool
    ) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/projects?vcsInfo={vcs_info}"
            f"&includeDisabled={include_disabled}"
        ).resolve()
        return self._http_client.get_json(url)

    def project_create(self, workspace_id: str, name: str) -> Dict[str, Any]:
        url = (self._base_uri / f"api/workspaces/{workspace_id}/projects").resolve()
        return self._http_client.post_json(url, body={"new": {"name": name}})

    def dependencies_get(
        self, workspace_id: str, vcs_info: bool, include_disabled: bool
    ) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/dependencies?vcsInfo={vcs_info}"
            f"&includeDisabled={include_disabled}"
        ).resolve()
        return self._http_client.get_json(url)

    def workspace_export_setup(
        self,
        workspace_id: str,
        publish: bool,
        class_path: Optional[str] = None,
        access_settings: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        url = (self._base_uri / "api/workspace-exports").resolve()
        body = {"workspaceId": workspace_id, "publish": publish}
        if access_settings:
            body["access"] = access_settings
        if class_path:
            body["appMode"] = {"model": class_path}
        return self._http_client.post_json(url, body=body)

    def workspace_conversion_setup(
        self, workspace_id: str, backup_name: Optional[str]
    ) -> Dict[str, Any]:
        url = (self._base_uri / "api/workspace-conversions").resolve()
        backup_data = {"backup": {"name": backup_name}} if backup_name else {}
        body: Dict[str, Any] = {"data": {"workspaceId": workspace_id, **backup_data}}
        return self._http_client.post_json(url, body=body)

    def get_workspace_conversion_status(self, location: str) -> Dict[str, Any]:
        url = (self._base_uri / location).resolve()
        return self._http_client.get_json(url)

    def fmus_get(self, workspace_id: str) -> Dict[str, Any]:
        url = (
            self._base_uri / f"api/workspaces/{workspace_id}/model-executables"
        ).resolve()
        return self._http_client.get_json(url)

    def fmu_get(self, workspace_id: str, fmu_id: str) -> Dict[str, Any]:
        url = (
            self._base_uri / f"api/workspaces/{workspace_id}/model-executables/{fmu_id}"
        ).resolve()
        return self._http_client.get_json(url)

    def fmu_download(self, workspace_id: str, fmu_id: str) -> bytes:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/model-executables/{fmu_id}/binary"
        ).resolve()
        return self._http_client.get_zip(url)

    def experiments_get(
        self, workspace_id: str, class_path: Optional[str] = None
    ) -> Dict[str, Any]:
        class_path_query = f"?classPath={class_path}" if class_path else ""
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments{class_path_query}"
        ).resolve()
        return self._http_client.get_json(
            url, headers={"Accept": self._experiment_schema}
        )

    def experiment_get(self, workspace_id: str, experiment_id: str) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}"
        ).resolve()
        return self._http_client.get_json(
            url, headers={"Accept": self._experiment_schema}
        )

    def experiment_create(
        self,
        workspace_id: str,
        definition: Dict[str, Any],
        user_data: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        url = (self._base_uri / f"api/workspaces/{workspace_id}/experiments").resolve()
        body = {
            **definition,
            **({"userData": user_data} if user_data is not None else {}),
        }
        return self._http_client.post_json(url, body=body)

    def shared_definition_get(
        self, workspace_id: str, strict: bool = False
    ) -> Dict[str, Any]:
        url = (
            self._base_uri / f"api/workspaces/{workspace_id}/"
            f"sharing-definition?strict={'true' if strict else 'false'}"
        ).resolve()
        return self._http_client.get_json(url)

    def import_from_zip(self, path_to_workspace: str) -> Dict[str, Any]:
        url = (self._base_uri / "api/workspace-imports").resolve()
        with open(path_to_workspace, "rb") as f:
            return self._http_client.post_json(url, files={"file": f})

    def import_from_shared_definition(
        self,
        shared_definition: Dict[str, Any],
        selected_matchings: Optional[List[Dict[str, Any]]] = None,
    ) -> Dict[str, Any]:
        if selected_matchings:
            shared_definition = {
                **shared_definition,
                **{"selectedMatchings": {"entries": selected_matchings}},
            }
        url = (self._base_uri / "api/workspace-imports").resolve()
        return self._http_client.post_json(url, body=shared_definition)

    def import_from_cloud(
        self, sharing_id: str, overwrite_workspace_id: Optional[str] = None
    ) -> Dict[str, Any]:
        url = (self._base_uri / "api/workspace-imports").resolve()
        payload: Dict[str, Any] = {"id": sharing_id}
        if overwrite_workspace_id:
            payload["update"] = {"workspaceId": overwrite_workspace_id}
        return self._http_client.post_json(
            url,
            headers={
                "Content-type": "application/vnd.impact.published-workspace.v1+json"
            },
            body=payload,
        )

    def get_project_matchings(
        self, shared_definition: Dict[str, Any]
    ) -> Dict[str, Any]:
        url = (self._base_uri / "api/workspace-imports-matchings").resolve()
        return self._http_client.post_json(url, body=shared_definition)

    def import_project_from_zip(
        self, workspace_id: str, path_to_project: str
    ) -> Dict[str, Any]:
        url = (
            self._base_uri / f"api/workspaces/{workspace_id}/project-imports"
        ).resolve()
        with open(path_to_project, "rb") as f:
            return self._http_client.post_json(url, files={"file": f})

    def import_dependency_from_zip(
        self, workspace_id: str, path_to_project: str
    ) -> Dict[str, Any]:
        url = (
            self._base_uri / f"api/workspaces/{workspace_id}/dependency-imports"
        ).resolve()
        with open(path_to_project, "rb") as f:
            return self._http_client.post_json(url, files={"file": f})

    def get_published_workspaces(
        self,
        name: str = "",
        first: int = 0,
        maximum: int = 20,
        has_data: bool = False,
        owner_username: str = "",
        type: Optional[str] = None,
        group_name: Optional[str] = None,
    ) -> Dict[str, Any]:
        query = {}
        if name:
            query["workspaceName"] = name
        if has_data:
            query["hasData"] = str(has_data)
        if first > 0:
            query["first"] = str(first)
        if maximum >= 0:
            query["max"] = str(maximum)
        if owner_username:
            query["ownerUsername"] = owner_username
        if type:
            query["type"] = type
        if group_name:
            query["groupName"] = group_name
        query_args = "&".join(f"{key}={value}" for key, value in query.items())
        url = (self._base_uri / f"api/published-workspaces?{query_args}").resolve()
        resp = self._http_client.get_json_response(url)
        return resp.data

    def get_published_workspaces_by_kind(
        self,
        kind: str = "",
        first: int = 0,
        maximum: int = 20,
    ) -> Dict[str, Any]:
        query = {}
        if kind:
            query["kind"] = kind
        if first > 0:
            query["first"] = str(first)
        if maximum >= 0:
            query["max"] = str(maximum)
        query_args = "&".join(f"{key}={value}" for key, value in query.items())
        url = (
            self._base_uri / f"/api/published-workspaces/access/users?{query_args}"
        ).resolve()
        resp = self._http_client.get_json_response(url)
        return resp.data

    def get_published_workspace(self, sharing_id: str) -> Dict[str, Any]:
        url = (self._base_uri / f"api/published-workspaces/{sharing_id}").resolve()
        resp = self._http_client.get_json_response(url)
        return resp.data

    def get_published_workspace_acl(self, sharing_id: str) -> Dict[str, Any]:
        url = (
            self._base_uri / f"api/published-workspaces/{sharing_id}/access"
        ).resolve()
        resp = self._http_client.get_json_response(url)
        return resp.data

    def rename_published_workspace(self, sharing_id: str, workspace_name: str) -> None:
        url = (self._base_uri / f"api/published-workspaces/{sharing_id}").resolve()
        self._http_client.patch_json_no_response_body(
            url, body={"workspaceName": workspace_name}
        )

    def _user_access_request(
        self, operation: str, sharing_id: str, username: Optional[str] = None
    ) -> None:
        url = (
            self._base_uri / f"api/published-workspaces/{sharing_id}/access/users"
        ).resolve()
        body = {"operation": operation}
        if username:
            body["requesterUsername"] = username
        self._http_client.patch_json_no_response_body(url, body=body)

    def _group_access_request(
        self, operation: str, sharing_id: str, group_name: Optional[str] = None
    ) -> None:
        url = (
            self._base_uri / f"api/published-workspaces/{sharing_id}/access/group"
        ).resolve()
        body = {"operation": operation}
        if group_name:
            body["groupName"] = group_name
        self._http_client.patch_json_no_response_body(url, body=body)

    def _community_access_request(self, operation: str, sharing_id: str) -> None:
        url = (
            self._base_uri / f"api/published-workspaces/{sharing_id}/access/community"
        ).resolve()
        body = {"operation": operation}
        self._http_client.patch_json_no_response_body(url, body=body)

    def grant_group_access(
        self, sharing_id: str, group_name: Optional[str] = None
    ) -> None:
        self._group_access_request("grant", sharing_id, group_name)

    def revoke_group_access(self, sharing_id: str, group_name: str) -> None:
        self._group_access_request("revoke", sharing_id, group_name)

    def request_user_access(self, sharing_id: str) -> None:
        self._user_access_request("request", sharing_id)

    def grant_user_access(self, sharing_id: str, username: str) -> None:
        self._user_access_request("grant", sharing_id, username)

    def revoke_user_access(self, sharing_id: str, username: str) -> None:
        self._user_access_request("revoke", sharing_id, username)

    def grant_community_access(self, sharing_id: str) -> None:
        self._community_access_request("grant", sharing_id)

    def revoke_community_access(self, sharing_id: str) -> None:
        self._community_access_request("revoke", sharing_id)

    def delete_published_workspace(self, sharing_id: str) -> None:
        url = (self._base_uri / f"api/published-workspaces/{sharing_id}").resolve()
        self._http_client.delete_json(url)

    def update_workspace(
        self, workspace_id: str, data: Dict[str, Any]
    ) -> Dict[str, Any]:
        url = (self._base_uri / f"api/workspaces/{workspace_id}").resolve()
        return self._http_client.put_json(url, body=data)
